import numpy as np

from pypop7.optimizers.es.es import ES


class SSAES(ES):
    """Schwefel's Self-Adaptation Evolution Strategy (SSAES).

    .. note:: `SSAES` adapts all the *individual* step-sizes on-the-fly, proposed by Schwefel. Since it needs the
       *relatively large* populations (e.g. larger than number of dimensionality) for reliable self-adaptation,
       `SSAES` easily suffers from *very slow* convergence for large-scale black-box optimization (LSBBO). Therefore,
       it is **highly recommended** to first attempt other more advanced ES variants (e.g. `LMCMA`, `LMMAES`) for
       LSBBO. Here we include it only for *benchmarking* and *theoretical* purpose.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'          - initial global step-size, aka mutation strength (`float`),
                * 'mean'           - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals'  - number of offspring, aka offspring population size (`int`, default:
                  `5*problem.get('ndim_problem')`),
                * 'n_parents'      - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/4)`),
                * 'lr_sigma'       - learning rate of global step-size adaptation (`float`, default:
                  `1.0/np.sqrt(self.ndim_problem)`),
                * 'lr_axis_sigmas' - learning rate of individual step-sizes adaptation (`float`, default:
                  `1.0/np.power(self.ndim_problem, 1.0/4.0)`).

    Examples
    --------
    Use the `ES` optimizer `SSAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.ssaes import SSAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> ssaes = SSAES(problem, options)  # initialize the optimizer class
       >>> results = ssaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SSAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SSAES: 5000, 0.13131869620062903

    Attributes
    ----------
    lr_axis_sigmas : `float`
                     learning rate of individual step-sizes adaptation.
    lr_sigma       : `float`
                     learning rate of global step-size adaptation.
    mean           : `array_like`
                     initial mean of Gaussian search distribution.
    n_individuals  : `int`
                     number of offspring, aka offspring population size.
    n_parents      : `int`
                     number of parents, aka parental population size.
    sigma          : `float`
                     final global step-size, aka mutation strength.

    References
    ----------
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44
    """
    def __init__(self, problem, options):
        if options.get('n_individuals') is None:
            options['n_individuals'] = 5*problem.get('ndim_problem')
        if options.get('n_parents') is None:
            options['n_parents'] = int(options['n_individuals']/4)
        ES.__init__(self, problem, options)
        if self.lr_sigma is None:
            self.lr_sigma = 1.0/np.sqrt(self.ndim_problem)  # learning rate of global step-size adaptation
        assert self.lr_sigma > 0, f'`self.lr_sigma` = {self.lr_sigma}, but should > 0.'
        # set learning rate of individual step-sizes adaptation
        self.lr_axis_sigmas = options.get('lr_axis_sigmas', 1.0/np.power(self.ndim_problem, 1.0/4.0))
        self._axis_sigmas = self.sigma*np.ones((self.ndim_problem,))  # individual step-sizes

    def initialize(self):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean()  # mean of Gaussian search distribution
        sigmas = np.empty((self.n_individuals, self.ndim_problem))  # individual step-sizes for all offspring
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return x, mean, sigmas, y

    def iterate(self, x=None, mean=None, sigmas=None, y=None, args=None):
        # sample offspring population
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, sigmas, y
            sigma = self.lr_sigma*self.rng_optimization.standard_normal()
            axis_sigmas = self.lr_axis_sigmas*self.rng_optimization.standard_normal((self.ndim_problem,))
            sigmas[k] = self._axis_sigmas*np.exp(axis_sigmas)*np.exp(sigma)
            x[k] = mean + sigmas[k]*self.rng_optimization.standard_normal((self.ndim_problem,))
            y[k] = self._evaluate_fitness(x[k], args)
        return x, sigmas, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, sigmas, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, sigmas, y = self.iterate(x, mean, sigmas, y, args)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            order = np.argsort(y)[:self.n_parents]
            self._axis_sigmas = np.mean(sigmas[order], axis=0)
            mean = np.mean(x[order], axis=0)
            self._print_verbose_info(y)
            self._n_generations += 1
        results = self._collect_results(fitness, mean)
        results['_axis_sigmas'] = self._axis_sigmas
        return results



"""Repeat the following paper for `IPSO`:
    De Oca, M.A.M., Stutzle, T., Van den Enden, K. and Dorigo, M., 2010.
    Incremental social learning in particle swarms.
    IEEE Transactions on Systems, Man, and Cybernetics, Part B (Cybernetics), 41(2), pp.368-384.
    https://ieeexplore.ieee.org/document/5582312
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import ackley
from pypop7.optimizers.pso.ipsols import IPSOLS as Solver

if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 100
    for f in [ackley]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -32.768 * np.ones((ndim_problem,)),
                   'upper_boundary': 32.768 * np.ones((ndim_problem,))}
        options = {'seed_rng': 0,  # not given in the original paper
                   'max_function_evaluations': 1e6,
                   'verbose': 1e3}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)      # 1.4559745853404138e-08 vs 5.91e-10 (from original paper)
        print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np

from pypop7.benchmarks import shifted_functions as sf
from pypop7.benchmarks.shifted_functions import generate_shift_vector


def generate_shift_vectors(functions=None, ndims=None, seed=None):
    if functions is None:
        functions = [sf.sphere, sf.cigar, sf.discus, sf.cigar_discus, sf.ellipsoid,
                     sf.different_powers, sf.schwefel221, sf.step, sf.rosenbrock, sf.schwefel12]
    if ndims is None:
        ndims = [2, 10, 100, 200, 1000, 2000]
    if seed is None:
        seed = 2022

    rng = np.random.default_rng(seed)
    seeds = rng.integers(np.iinfo(np.int64).max, size=(len(functions), len(ndims)))

    for i, f in enumerate(functions):
        for j, d in enumerate(ndims):
            generate_shift_vector(f, d, -9.5, 9.5, seeds[i, j])


if __name__ == '__main__':
    generate_shift_vectors()



"""Repeat experiments on Fig10(a)(b) from the following paper:
    W. Dong, T. Chen, P. Tino, X. Yao
    Scaling Up Estimation of Distribution Algorithms for Continuous Optimization
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 17, NO. 6, DECEMBER 2013
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=6461934
"""
import sys
import pickle
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import schwefel12, schwefel221
from pypop7.optimizers.core import optimizer
from pypop7.optimizers.eda.edamcc import EDAMCC as Solver

sys.modules['optimizer'] = optimizer
sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function):
    plt.figure()
    result = read_pickle(function, 100)
    plt.yscale('log')
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='r')
    plt.xlim([0, 1e6])
    plt.ylim([1e-30, 1e10])
    plt.xticks([0, 2e5, 4e5, 6e5, 8e5, 1e6])
    plt.yticks([1e-30, 1e-20, 1e-10, 1e0, 1e10])
    plt.xlabel("#evals")
    plt.ylabel("F(x)-F(x*)")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    ndim_problem = 100
    for f in [schwefel221, schwefel12]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -10 * np.ones((ndim_problem,)),
                   'upper_boundary': 10 * np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 1e6,
                   'fitness_threshold': 1e-10,
                   'seed_rng': 0,
                   'theta': 0.3,
                   'c': 20,
                   'n_individuals': 200,
                   'verbose_frequency': 20,
                   'record_fitness': True,
                   'record_fitness_frequency': 1}
        if f == schwefel221:
            problem['lower_boundary'] = -100 * np.ones((ndim_problem,))
            problem['upper_boundary'] = 100 * np.ones((ndim_problem,))
        solver = Solver(problem, options)
        results = solver.optimize()
        write_pickle(f.__name__, ndim_problem, results)
        plot(f.__name__)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class GA(Optimizer):
    """Genetic Algorithm (GA).

    This is the **base** (abstract) class for all `GA` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_individuals' - population size (`int`, default: `100`).

    Attributes
    ----------
    n_individuals : `int`
                    population size.

    Methods
    -------

    References
    ----------
    Whitley, D., 2019.
    Next generation genetic algorithms: A user’s guide and tutorial.
    In Handbook of Metaheuristics (pp. 245-274). Springer, Cham.
    https://link.springer.com/chapter/10.1007/978-3-319-91086-4_8

    De Jong, K.A., 2006.
    Evolutionary computation: A unified approach.
    MIT Press.
    https://mitpress.mit.edu/9780262041942/evolutionary-computation/

    Levine, D., 1997.
    Commentary—Genetic algorithms: A practitioner's view.
    INFORMS Journal on Computing, 9(3), pp.256-259.
    https://pubsonline.informs.org/doi/10.1287/ijoc.9.3.256

    Goldberg, D.E., 1994.
    Genetic and evolutionary algorithms come of age.
    Communications of the ACM, 37(3), pp.113-120.
    https://dl.acm.org/doi/10.1145/175247.175259

    De Jong, K.A., 1993.
    Are genetic algorithms function optimizer?.
    Foundations of Genetic Algorithms, pp.5-17.
    https://www.sciencedirect.com/science/article/pii/B9780080948324500064

    Forrest, S., 1993.
    Genetic algorithms: Principles of natural selection applied to computation.
    Science, 261(5123), pp.872-878.
    https://www.science.org/doi/10.1126/science.8346439

    Mitchell, M., Holland, J. and Forrest, S., 1993.
    When will a genetic algorithm outperform hill climbing.
    Advances in Neural Information Processing Systems (pp. 51-58).
    https://proceedings.neurips.cc/paper/1993/hash/ab88b15733f543179858600245108dd8-Abstract.html

    Holland, J.H., 1992.
    Genetic algorithms.
    Scientific American, 267(1), pp.66-73.
    https://www.scientificamerican.com/article/genetic-algorithms/

    Goldberg, D.E. and Holland, J.H., 1988.
    Genetic algorithms and machine learning.
    Machine Learning, 3(2), pp.95-99.
    https://link.springer.com/article/10.1023/A:1022602019183

    Holland, J.H., 1973.
    Genetic algorithms and the optimal allocation of trials.
    SIAM Journal on Computing, 2(2), pp.88-105.
    https://epubs.siam.org/doi/10.1137/0202009

    Holland, J.H., 1962.
    Outline for a logical theory of adaptive systems.
    Journal of the ACM, 9(3), pp.297-314.
    https://dl.acm.org/doi/10.1145/321127.321128
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # population size
            self.n_individuals = 100
        self._n_generations = 0

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness):
        results = Optimizer._collect_results(self, fitness)
        results['_n_generations'] = self._n_generations
        return results



from pypop7.optimizers.ga.ga import GA
from pypop7.optimizers.ga.genitor import GENITOR
from pypop7.optimizers.ga.g3pcx import G3PCX
from pypop7.optimizers.ga.gl25 import GL25
from pypop7.optimizers.ga.asga import ASGA


__all__ = [GA, GENITOR, G3PCX, GL25, ASGA]



"""A short and simple example experiment of coco benchmark

This code is referenced from following library:
https://github.com/numbbo/coco

In this program we have implement our maes algorithm to the coco library

"""
from __future__ import division, print_function
import cocoex, cocopp  # experimentation and post-processing modules
from numpy.random import rand  # for randomised restarts
from pypop7.optimizers.es.maes import MAES as Solver
import os, webbrowser  # to show post-processed results in the browser

### input
suite_name = "bbob"
output_folder = "coco-benchmarking-maes-fmin"
budget_multiplier = 1  # increase to 10, 100, ...
def fmin(fun, x0, disp=False):
    problem = {
        'fitness_function': fun,
        'ndim_problem': fun.dimension,
        'lower_boundary': fun.lower_bounds,
        'upper_boundary': fun.upper_bounds
    }
    options = {
        'fitness_threshold': 1e-10,
        'max_function_evaluations': 1e1,
        'seed_rng': 2022,
        'x': x0,
        'sigma': 1.0,
        'verbose': disp
    }
    solver = Solver(problem, options)
    results = solver.optimize()
    return results['best_so_far_x']

### prepare
suite = cocoex.Suite(suite_name, "", "")
observer = cocoex.Observer(suite_name, "result_folder: " + output_folder)
minimal_print = cocoex.utilities.MiniPrint()

### go
for problem in suite:  # this loop will take several minutes or longer
    problem.observe_with(observer)  # generates the data for cocopp post-processing
    x0 = problem.initial_solution
    # apply restarts while neither the problem is solved nor the budget is exhausted
    while (problem.evaluations < problem.dimension * budget_multiplier
           and not problem.final_target_hit):
        result = fmin(problem, x0, disp=False)  # here we assume that `fmin` evaluates the final/returned solution
        x0 = problem.lower_bounds + ((rand(problem.dimension) + rand(problem.dimension)) *
                    (problem.upper_bounds - problem.lower_bounds) / 2)
    minimal_print(problem, final=problem.index == len(suite) - 1)

### post-process data
cocopp.main(observer.result_folder)  # re-run folders look like "...-001" etc
webbrowser.open("file://" + os.getcwd() + "/ppdata/index.html")


"""Repeat the following paper for `LMMAES`:
    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/lmmaes

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `LMMAES` could be **well-documented** (*at least partly*).
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere, cigar, ellipsoid, rosenbrock
from pypop7.optimizers.es.lmmaes import LMMAES as Solver


def plot(function):
    ndim = [128, 256, 512, 1024, 2048, 4096, 8192]
    colors = ['r', 'orange', 'y', 'limegreen', 'cyan', 'b', 'purple']
    plt.figure()
    for k in range(len(ndim)):
        result = pickle.load(open(function + '_' + str(ndim[k]) + '.pickle', 'rb'))
        plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color=colors[k],
                 label=str(ndim), linestyle='dashed')
    plt.yscale('log')
    plt.xscale('log')
    if function == 'sphere':
        plt.xlim([1e0, 1e6])
        plt.ylim([1e-10, 1e6])
        plt.xticks([1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6])
        plt.yticks(ticks=[1e-10, 1e-5, 1e0, 1e5])
    if function == 'cigar':
        plt.xlim([1e2, 1e8])
        plt.ylim([1e-10, 1e11])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
        plt.yticks(ticks=[1e-10, 1e-5, 1e0, 1e5, 1e10])
    if function == 'ellipsoid':
        plt.xlim([1e3, 1e9])
        plt.ylim([1e-10, 1e11])
        plt.xticks([1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9])
        plt.yticks(ticks=[1e-10, 1e-5, 1e0, 1e5, 1e10])
    if function == 'rosenbrock':
        plt.xlim([1e3, 1e9])
        plt.ylim([1e-10, 1e9])
        plt.xticks([1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9])
        plt.yticks(ticks=[1e-10, 1e-5, 1e0, 1e5, 1e10])
    plt.xlabel("number of function evaluations")
    plt.ylabel("objective")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    for f in [sphere, cigar, ellipsoid, rosenbrock]:
        for d in [128, 256, 512, 1024, 2048, 4096, 8192]:
            problem = {'fitness_function': f,
                       'ndim_problem': d}
            options = {'fitness_threshold': 1e-10,
                       'seed_rng': 0,  # undefined in the original paper
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 3,
                       'verbose': 5000,
                       'saving_fitness': 100,
                       'is_restart': False}
            solver = Solver(problem, options)
            results = solver.optimize()
            pickle.dump(results, open(f.__name__ + '_' + str(d) + '.pickle', 'wb'))
        plot(f.__name__)



from pypop7.optimizers.ds.ds import DS
from pypop7.optimizers.ds.cs import CS
from pypop7.optimizers.ds.hj import HJ
from pypop7.optimizers.ds.nm import NM
from pypop7.optimizers.ds.gps import GPS


__all__ = [DS, CS, HJ, NM, GPS]



"""This code only tests whether the corresponding optimizer could run or not.
  In other words, it *cannot* validate the correctness regarding its working procedure.
  For the correctness validation of programming, refer to the following link:
    https://github.com/Evolutionary-Intelligence/pypop/wiki/Correctness-Validation-of-Optimizers.md
"""
import unittest
import time

import numpy as np

from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock, rastrigin
from pypop7.optimizers.pso.spso import SPSO as Solver


class TestSPSO(unittest.TestCase):
    def test_optimize(self):
        start_run = time.time()
        ndim_problem = 1000
        for f in [ellipsoid, rosenbrock, rastrigin]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': ndim_problem,
                       'lower_boundary': -5 * np.ones((ndim_problem,)),
                       'upper_boundary': 5 * np.ones((ndim_problem,))}
            options = {'max_function_evaluations': 2e6,
                       'fitness_threshold': 1e-10,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'saving_fitness': 200000,
                       'verbose': 10000}
            solver = Solver(problem, options)
            results = solver.optimize()
            print(results)
            print('*** Runtime: {:7.5e}'.format(time.time() - start_run))


if __name__ == '__main__':
    unittest.main()



"""Repeat the following library for `COSYNE`:
    https://evotorch.ai/

    The referenced paper can be found by:
    F. Gomez, J. Schmidhuber, R. Miikkulainen
    Accelerated Neural Evolution through Cooperatively Coevolved Synapses
    https://jmlr.org/papers/v9/gomez08a.html

    Luckily our code could repeat the performance of the EvoTroch *well*.
    Therefore, we argue that the repeatability of `COSYNE` could be **well-documented**.
"""
import torch
from evotorch import Problem
from evotorch.algorithms import ga
from pypop7.benchmarks.base_functions import ellipsoid
from evotorch.logging import StdOutLogger


def myellipsoid(x: torch.Tensor) -> torch.Tensor:
    y = []
    for i in range(len(x)):
        y.append(ellipsoid(x[i].numpy()))
    return torch.tensor(y)


problem = Problem(
    "min",
    myellipsoid,
    initial_bounds=(-5.0, 5.0),
    solution_length=10,
    vectorized=True
)
searcher = ga.Cosyne(problem, popsize=20, tournament_size=2, mutation_probability=0.3, mutation_stdev=1.0)
_ = StdOutLogger(searcher, interval=50)

searcher.run(num_generations=3000)



"""Repeat the following paper for `OPOC2009`:
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 2 for details.)

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/opoc2009

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `OPOC2009` could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import ellipsoid, cigar
from pypop7.optimizers.es.opoc2009 import OPOC2009 as Solver


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    # plot fig.4
    d = 20
    problem = {'fitness_function': ellipsoid,
               'ndim_problem': d,
               'lower_boundary': 0.1 * np.ones((d,)),
               'upper_boundary': 0.3 * np.ones((d,))}
    options = {'fitness_threshold': 1e-250,
               'max_function_evaluations': 6e4,
               'seed_rng': 0,
               'sigma': 0.3,  # not given in the original paper
               'saving_fitness': 1,
               'is_restart': False}
    solver = Solver(problem, options)
    results = solver.optimize()
    pickle.dump(results, open(ellipsoid.__name__ + '_' + str(d) + '.pickle', 'wb'))
    plt.figure()
    result = pickle.load(open(ellipsoid.__name__ + '_' + str(d) + '.pickle', 'rb'))
    plt.xlim([0, 1e5])
    plt.xticks([0, 2e4, 4e4, 6e4, 8e4, 1e5])
    plt.xlabel("objective function evaluations")
    plt.yscale('log')
    plt.ylim([1e-250, 1e50])
    plt.yticks([1e-250, 1e-200, 1e-150, 1e-100, 1e-50, 1e0, 1e50])
    plt.ylabel("objective function value")
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='green')
    plt.title(ellipsoid.__name__.capitalize())
    plt.show()
    # plot fig.5
    evaluations = []
    dims = [5, 10, 20, 40, 80, 160, 320]
    for d in dims:
        problem = {'fitness_function': cigar,
                   'ndim_problem': d,
                   'lower_boundary': 0.1 * np.ones((d,)),
                   'upper_boundary': 0.3 * np.ones((d,))}
        options = {'fitness_threshold': 1e-15,
                   'max_function_evaluations': 1e6,
                   'seed_rng': 0,
                   'sigma': 0.3,  # not given in the original paper
                   'saving_fitness': 1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        evaluations.append(results['n_function_evaluations'])
    plt.figure()
    plt.xscale('log')
    plt.xlim([1, 5e2])
    plt.xticks([10, 100])
    plt.xlabel('search space dimension n')
    plt.yscale('log')
    plt.ylim([1e3, 1e7])
    plt.yticks([1e3, 1e4, 1e5, 1e6, 1e7])
    plt.ylabel('objective function evaluations')
    plt.plot(dims, evaluations, 'o--', color='green')
    plt.title(cigar.__name__.capitalize())
    plt.show()



import numpy as np
import numba as nb

from pypop7.optimizers.es.es import ES


@nb.jit(nopython=True)
def cholesky_update(rm, z, downdate):
    # https://github.com/scipy/scipy/blob/d20f92fce9f1956bfa65365feeec39621a071932/
    #     scipy/linalg/_decomp_cholesky_update.py
    rm, z, alpha, beta = rm.T, z, np.empty_like(z), np.empty_like(z)
    alpha[-1], beta[-1] = 1., 1.
    sign = -1 if downdate else 1
    for r in range(len(z)):
        a = z[r]/rm[r, r]
        alpha[r] = alpha[r - 1] + sign*np.power(a, 2)
        beta[r] = np.sqrt(alpha[r])
        z[r + 1:] -= a*rm[r, r + 1:]
        rm[r, r:] *= beta[r]/beta[r - 1]
        rm[r, r + 1:] += sign*a/(beta[r]*beta[r - 1])*z[r + 1:]
    return rm.T


class OPOA2015(ES):
    """(1+1)-Active-CMA-ES 2015 (OPOA2015).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`).

    Examples
    --------
    Use the `ES` optimizer `OPOA2015` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.opoa2015 import OPOA2015
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> opoa2015 = OPOA2015(problem, options)  # initialize the optimizer class
       >>> results = opoa2015.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"OPOA2015: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       OPOA2015: 5000, 6.710440903901067e-16

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mrxu4suj>`_ for more details.

    References
    ----------
    Krause, O. and Igel, C., 2015, January.
    A more efficient rank-one covariance matrix update for evolution strategies.
    In Proceedings of ACM Conference on Foundations of Genetic Algorithms (pp. 129-136).
    https://dl.acm.org/doi/abs/10.1145/2725494.2725496
    """
    def __init__(self, problem, options):
        options['n_individuals'] = 1  # mandatory setting for OPOA2015
        options['n_parents'] = 1  # mandatory setting for OPOA2015
        ES.__init__(self, problem, options)
        if self.lr_sigma is None:
            self.lr_sigma = 1.0/(1.0 + self.ndim_problem/2.0)
        self.p_ts = options.get('p_ts', 2.0/11.0)
        self.c_p = options.get('c_p', 1.0/12.0)
        self.c_c = options.get('c_c', 2.0/(self.ndim_problem + 2.0))
        self.c_cov = options.get('c_cov', 2.0/(np.power(self.ndim_problem, 2) + 6.0))
        self.p_t = options.get('p_t', 0.44)
        self.c_m = options.get('c_m', 0.4/(np.power(self.ndim_problem, 1.6) + 1.0))
        self.k = options.get('k', 5)
        self._ancestors = []
        self._c_cf = 1.0 - self.c_cov + self.c_cov*self.c_c*(2.0 - self.c_c)

    def initialize(self, args=None, is_restart=False):
        mean = self._initialize_mean(is_restart)
        y = self._evaluate_fitness(mean, args)  # fitness
        cf = np.diag(np.ones(self.ndim_problem,))  # Cholesky factorization
        best_so_far_y, p_s = np.copy(y), self.p_ts
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        return mean, y, cf, best_so_far_y, p_s, p_c

    def _cholesky_update(self, cf=None, alpha=None, beta=None, v=None):  # triangular rank-one update
        assert self.ndim_problem == v.size
        if beta < 0:
            downdate, beta = True, -beta
        else:
            downdate = False
        return cholesky_update(np.sqrt(alpha)*cf, np.sqrt(beta)*v, downdate)

    def iterate(self, args=None, mean=None, cf=None, best_so_far_y=None, p_s=None, p_c=None):
        # sample and evaluate only one offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        cf_z = np.dot(cf, z)
        x = mean + self.sigma*cf_z
        y = self._evaluate_fitness(x, args)
        if y <= best_so_far_y:
            self._ancestors.append(y)
            mean, best_so_far_y = x, y
            p_s = (1.0 - self.c_p)*p_s + self.c_p
            is_better = True
        else:
            p_s *= 1.0 - self.c_p
            is_better = False
        self.sigma *= np.exp(self.lr_sigma*(p_s - self.p_ts)/(1.0 - self.p_ts))
        if p_s >= self.p_t:
            p_c *= 1.0 - self.c_c
            cf = self._cholesky_update(cf, self._c_cf, self.c_cov, p_c)
        elif is_better:
            p_c = (1.0 - self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c))*cf_z
            cf = self._cholesky_update(cf, 1.0 - self.c_cov, self.c_cov, p_c)
        elif len(self._ancestors) >= self.k and y > self._ancestors[-self.k]:
            del self._ancestors[0]
            c_m = np.minimum(self.c_m, 1.0/(2.0*np.dot(z, z) - 1.0))
            cf = self._cholesky_update(cf, 1.0 + c_m, -c_m, cf_z)
        return mean, cf, best_so_far_y, p_s, p_c, y

    def restart_reinitialize(self, args=None, mean=None, y=None, cf=None,
                             best_so_far_y=None, p_s=None, p_c=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, cf, best_so_far_y, p_s, p_c = self.initialize(args, True)
            if self.saving_fitness:
                fitness.append(y)
            self._fitness_list = [best_so_far_y]
            self._ancestors = []
            self._n_generations = 0
            if self.verbose:
                print(' ....... restart .......')
            self._print_verbose_info(y)
        return mean, y, cf, best_so_far_y, p_s, p_c

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, cf, best_so_far_y, p_s, p_c = self.initialize(args)
        if self.saving_fitness:
            fitness.append(y)
        self._print_verbose_info(y)
        while True:
            mean, cf, best_so_far_y, p_s, p_c, y = self.iterate(
                args, mean, cf, best_so_far_y, p_s, p_c)
            if self.saving_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, cf, best_so_far_y, p_s, p_c = self.restart_reinitialize(
                    args, mean, y, cf, best_so_far_y, p_s, p_c, fitness)
        return self._collect_results(fitness, mean)



import numpy as np

from pypop7.optimizers.es.opoc2009 import OPOC2009


class OPOA2010(OPOC2009):
    """(1+1)-Active-CMA-ES 2010 (OPOA2010).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`).

    Examples
    --------
    Use the `ES` optimizer `OPOA2010` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.opoa2010 import OPOA2010
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> opoa2010 = OPOA2010(problem, options)  # initialize the optimizer class
       >>> results = opoa2010.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"OPOA2010: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       OPOA2010: 5000, 7.908973845675895e-06

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/26tad82p>`_ for more details.

    References
    ----------
    Arnold, D.V. and Hansen, N., 2010, July.
    Active covariance matrix adaptation for the (1+1)-CMA-ES.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 385-392). ACM.
    https://dl.acm.org/doi/abs/10.1145/1830483.1830556
    """
    def __init__(self, problem, options):
        OPOC2009.__init__(self, problem, options)
        self.c_m = options.get('c_m', 0.4/(np.power(self.ndim_problem, 1.6) + 1.0))
        self.k = options.get('k', 5)
        self._ancestors = []

    def initialize(self, args=None, is_restart=False):
        mean, y, a, a_i, best_so_far_y, p_s, p_c = OPOC2009.initialize(self, args, is_restart)
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def iterate(self, args=None, mean=None, a=None, a_i=None, best_so_far_y=None, p_s=None, p_c=None):
        # sample and evaluate only one offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        x = mean + self.sigma*np.dot(a, z)
        y = self._evaluate_fitness(x, args)
        l_s = 1 if y <= best_so_far_y else 0
        p_s = (1.0 - self.c_p)*p_s + self.c_p*l_s
        self.sigma *= np.exp(self.lr_sigma*(p_s - self.p_ts)/(1.0 - self.p_ts))
        if y <= best_so_far_y:
            self._ancestors.append(y)
            mean, best_so_far_y = x, y
            if p_s < self.p_t:
                p_c = (1.0 - self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c))*np.dot(a, z)
                alpha = 1.0 - self.c_cov
            else:
                p_c *= 1.0 - self.c_c
                alpha = 1.0 - self.c_cov + self.c_cov*self.c_c*(2.0 - self.c_c)
            w = np.dot(a_i, p_c)
            w_power = np.dot(w, w)
            alpha = np.sqrt(alpha)
            beta = alpha/w_power*(np.sqrt(1.0 + self.c_cov/(1.0 - self.c_cov)*w_power) - 1.0)
            a = alpha*a + beta*np.dot(p_c[:, np.newaxis], w[np.newaxis, :])
            a_i = 1.0/alpha*a_i - beta/(np.power(alpha, 2) + alpha*beta*w_power)*np.dot(
                w[:, np.newaxis], np.dot(w[np.newaxis, :], a_i))
        if len(self._ancestors) >= self.k and y > self._ancestors[-self.k]:
            del self._ancestors[0]
            z_power = np.dot(z, z)
            if 1.0 < self.c_m*(2.0*z_power - 1.0):
                c_m = 1.0/(2.0*z_power - 1.0)
            else:
                c_m = self.c_m
            alpha = np.sqrt(1.0 + c_m)
            beta = alpha/z_power*(np.sqrt(1.0 - self.c_m/(1.0 - self.c_m)*z_power) - 1.0)
            a = alpha*a + beta*np.dot(np.dot(a, z[:, np.newaxis]), z[np.newaxis, :])
            a_i = 1.0/alpha*a_i - beta/(np.power(alpha, 2) + alpha*beta*z_power)*np.dot(
                z[:, np.newaxis], np.dot(z[np.newaxis, :], a_i))
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def restart_initialize(self, args=None, mean=None, y=None, a=None, a_i=None,
                           best_so_far_y=None, p_s=None, p_c=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args, True)
            if self.saving_fitness:
                fitness.append(y)
            self._fitness_list = [best_so_far_y]
            self._ancestors = []
            self._n_generations = 0
            if self.verbose:
                print(' ....... restart .......')
            self._print_verbose_info(y)
        return mean, y, a, a_i, best_so_far_y, p_s, p_c



import numpy as np
from scipy.linalg import solve_triangular

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.opoa2015 import cholesky_update


class CCMAES2016(ES):
    """Cholesky-CMA-ES 2016 (CCMAES2016).

    References
    ----------
    Krause, O., Arbonès, D.R. and Igel, C., 2016.
    CMA-ES with optimal covariance update and storage complexity.
    Advances in Neural Information Processing Systems, 29, pp.370-378.
    https://proceedings.neurips.cc/paper/2016/hash/289dff07669d7a23de0ef88d2f7129e7-Abstract.html
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_s = options.get('c_s', self._set_c_s())
        self.d = options.get('d', self._set_d())
        self.c_c = options.get('c_c', self._set_c_c())
        self.c_1 = options.get('c_1', self._set_c_1())
        self.c_mu = options.get('c_mu', self._set_c_mu())

    def _set_c_s(self):
        return self._mu_eff/(self.ndim_problem + self._mu_eff)

    def _set_d(self):
        return 1.0 + np.sqrt(self._mu_eff/self.ndim_problem)

    def _set_c_c(self):
        return (4.0 + self._mu_eff/self.ndim_problem)/(
                self.ndim_problem + 4.0 + 2.0*self._mu_eff/self.ndim_problem)

    def _set_c_1(self):
        return 2.0/(np.power(self.ndim_problem, 2) + self._mu_eff)

    def _set_c_mu(self):
        return self._mu_eff/(np.power(self.ndim_problem, 2) + self._mu_eff)

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        a = np.diag(np.ones((self.ndim_problem,)))  # cholesky factor
        p_s = np.zeros((self.ndim_problem,))  # evolution path for CSA
        p_c = np.zeros((self.ndim_problem,))  # evolution path for CMA
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        if is_restart:
            self.c_s = self._set_c_s()
            self.d = self._set_d()
            self.c_c = self._set_c_c()
            self.c_1 = self._set_c_1()
            self.c_mu = self._set_c_mu()
        return x, mean, a, p_s, p_c, y

    def iterate(self, x=None, mean=None, a=None, y=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            x[k] = mean + self.sigma*np.dot(a, self.rng_optimization.standard_normal((self.ndim_problem,)))
            y[k] = self._evaluate_fitness(x[k])
        return x, y

    def _update_distribution(self, x=None, mean=None, a=None, p_s=None, p_c=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        mean_bak = np.dot(self._w, x[order])
        mean_diff = (mean_bak - mean)/self.sigma
        p_c = (1.0 - self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*mean_diff
        p_s = (1.0 - self.c_s)*p_s + np.sqrt(self.c_s*(2.0 - self.c_s)*self._mu_eff)*solve_triangular(
            a, mean_diff, lower=True)
        a *= np.sqrt(1.0 - self.c_1 - self.c_mu)
        a = cholesky_update(a, np.sqrt(self.c_1)*p_c, False)
        for i in range(self.n_parents):
            a = cholesky_update(a, np.sqrt(self.c_mu*self._w[i])*(x[order[i]] - mean)/self.sigma, False)
        self.sigma *= np.exp(self.c_s/self.d*(np.sqrt(np.dot(p_s, p_s))/self._e_chi - 1.0))
        mean = mean_bak
        return mean, a, p_s, p_c

    def restart_reinitialize(self, x=None, mean=None, a=None, p_s=None, p_c=None, y=None):
        if ES.restart_reinitialize(self):
            x, mean, a, p_s, p_c, y = self.initialize(True)
        return x, mean, a, p_s, p_c, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, a, p_s, p_c, y = self.initialize()
        while True:
            x, y = self.iterate(x, mean, a, y)  # sample and evaluate offspring population
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, a, p_s, p_c = self._update_distribution(x, mean, a, p_s, p_c, y)
            self._print_verbose_info(y)
            self._n_generations += 1
            if self.is_restart:
                x, mean, a, p_s, p_c, y = self.restart_reinitialize(x, mean, a, p_s, p_c, y)
        results = self._collect_results(fitness, mean)
        results['p_s'] = p_s
        results['p_c'] = p_c
        return results



"""Repeat the following paper for `NM`:
    Nelder, J.A. and Mead, R., 1965.
    A simplex method for function minimization.
    The Computer Journal, 7(4), pp.308-313.
    https://academic.oup.com/comjnl/article-abstract/7/4/308/354237

    Luckily our code could repeat the data generated by the other Python code *well*.
    Therefore, we argue that the repeatability of `NM` could be **well-documented**.



    You can run the following Python script (note that first install `pymoo` via `pip install pymoo`):

    from pymoo.algorithms.soo.nonconvex.nelder import NelderMead
    from pymoo.problems.single import Ackley
    from pymoo.optimize import minimize

    problem = Ackley(n_var=100)
    algorithm = NelderMead(init_delta=0.1)
    res = minimize(problem=problem, algorithm=algorithm, termination=('n_eval', 1e5), verbose=True, seed=0)
    print(res)
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import ackley
from pypop7.optimizers.ds.nm import NM as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 100
    for f in [ackley]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -32.768 * np.ones((ndim_problem,)),
                   'upper_boundary': 32.768 * np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 1e5,
                   'seed_rng': 0,
                   'sigma': 0.1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)  # 19.805781729365254 vs 1.958869E+01 (from pymoo)
        print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



"""Repeat the following paper for `ASGA`:
    Demo, N., Tezzele, M. and Rozza, G., 2021.
    A supervised learning approach involving active subspaces for an efficient genetic algorithm in
      high-dimensional optimization problems.
    SIAM Journal on Scientific Computing, 43(3), pp.B831-B853.
    https://epubs.siam.org/doi/10.1137/20M1345219

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `ASGA` could be **well-documented**.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, ackley
from pypop7.optimizers.ga.asga import ASGA


if __name__ == '__main__':
    ndim_problem = 15
    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2000 + 200 * 30,
               'n_initial_individuals': 2000,
               'n_individuals': 200,
               'seed_rng': 0,
               'verbose': 1}
    asga = ASGA(problem, options)
    results = asga.optimize()
    print(results)
    print(results['best_so_far_y'])  # 14.03062287458637

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -15 * np.ones((ndim_problem,)),
               'upper_boundary': 30 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2000 + 200 * 30,
               'n_initial_individuals': 2000,
               'n_individuals': 200,
               'seed_rng': 0,
               'verbose': 1}
    asga = ASGA(problem, options)
    results = asga.optimize()
    print(results)
    print(results['best_so_far_y'])  # 0.012111557779814763



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.rs.rs import RS


class BES(RS):
    """BErnoulli Smoothing (BES).

    .. note:: This is a **highly simplified** version of the original BES algorithm without noisy fitness evaluations.
       We leave the *noisy fitness evaluations* case for the future work.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of individuals/samples (`int`, default: `100`),
                * 'lr'            - learning rate (`float`, default: `0.001`),
                * 'c'             - factor of finite-difference gradient estimate (`float`, default: `0.1`).

    Examples
    --------
    Use the Random Search optimizer `BES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.rs.bes import BES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 100,
       ...            'lower_boundary': -2 * numpy.ones((100,)),
       ...            'upper_boundary': 2 * numpy.ones((100,))}
       >>> options = {'max_function_evaluations': 10000*101,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'n_individuals': 10,
       ...            'c': 0.1,
       ...            'lr': 0.000001}
       >>> bes = BES(problem, options)  # initialize the optimizer class
       >>> results = bes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"BES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       BES: 1010000, 133.79696876596637

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/37hke555>`_ for more details.

    Attributes
    ----------
    c             : `float`
                    factor of finite-difference gradient estimate.
    lr            : `float`
                    learning rate.
    n_individuals : `int`
                    number of individuals/samples.

    References
    ----------
    Gao, K. and Sener, O., 2022, June.
    Generalizing Gaussian Smoothing for Random Search.
    In International Conference on Machine Learning (pp. 7077-7101). PMLR.
    https://proceedings.mlr.press/v162/gao22f.html
    https://icml.cc/media/icml-2022/Slides/16434.pdf
    """
    def __init__(self, problem, options):
        RS.__init__(self, problem, options)
        self.n_individuals = options.get('n_individuals', 100)  # number of individuals (samples)
        self.lr = options.get('lr', 0.001)  # learning rate
        self.c = options.get('c', 0.1)  # factor of finite-difference gradient estimate
        self.verbose = options.get('verbose', 10)

    def initialize(self, is_restart=False):
        if is_restart or (self.x is None):
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        return x

    def iterate(self, args=None, x=None, fitness=None):  # for each iteration (generation)
        gradient = np.zeros((self.ndim_problem,))  # estimated gradient
        y_base = self._evaluate_fitness(x, args)  # for finite-difference gradient estimate
        if self.saving_fitness:
            fitness.append(y_base)
        self._print_verbose_info(y_base)
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x
            # set directional gradient based on Bernoulli distribution
            dg = self.rng_optimization.binomial(n=1, p=0.5, size=(self.ndim_problem,))
            dg = (dg - 0.5)/0.5
            y[i] = self._evaluate_fitness(x + self.c*dg, args)
            gradient += (y[i] - y_base)*dg
            if self.saving_fitness:
                fitness.append(y[i])
        gradient /= (self.c*self.n_individuals)
        x -= self.lr*gradient  # stochastic gradient descent (SGD)
        self._n_generations += 1
        self._print_verbose_info(y)
        return x

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        x = self.initialize()
        while not self._check_terminations():
            x = self.iterate(args, x, fitness)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.es.es import ES


class CCMAES2009(ES):
    """Cholesky-CMA-ES 2009 (CCMAES2009).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`).

    Examples
    --------
    Use the `ES` optimizer `CCMAES2009` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.ccmaes2009 import CCMAES2009
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> ccmaes2009 = CCMAES2009(problem, options)  # initialize the optimizer class
       >>> results = ccmaes2009.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CCMAES2009: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CCMAES2009: 5000, 4.423033214966621e-16

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/c5hreha9>`_ for more details.

    References
    ----------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 4 for details.)
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_s = options.get('c_s', self._set_c_s())
        self.d_s = options.get('d_s', self._set_d_s())
        self.c_c = options.get('c_c', 4.0/(self.ndim_problem + 4.0))
        self.c_cov = options.get('c_cov', 2.0/np.power(self.ndim_problem + np.sqrt(2.0), 2.0))

    def _set_c_s(self):
        return np.sqrt(self._mu_eff)/(np.sqrt(self.ndim_problem) + np.sqrt(self._mu_eff))

    def _set_d_s(self):
        return 1.0 + 2.0*np.maximum(0.0, np.sqrt((self._mu_eff - 1.0)/(self.ndim_problem + 1.0)) - 1.0) + self.c_s

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        a = np.diag(np.ones(self.ndim_problem,))  # Cholesky factors
        a_i = np.diag(np.ones(self.ndim_problem,))  # inverse of Cholesky factors
        p_s = np.zeros((self.ndim_problem,))  # evolution path for global step-size adaptation
        p_c = np.zeros((self.ndim_problem,))  # evolution path for covariance matrix adaptation
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        if is_restart:
            self.c_s, self.d_s = self._set_c_s(), self._set_d_s()
        return mean, z, x, a, a_i, p_s, p_c, y

    def iterate(self, z=None, x=None, mean=None, a=None, y=None, args=None):
        for k in range(self.n_individuals):  # Line 4
            if self._check_terminations():
                return z, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 5
            x[k] = mean + self.sigma*np.dot(a, z[k])  # Line 6
            y[k] = self._evaluate_fitness(x[k], args)
        return z, x, y

    def _update_distribution(self, z=None, x=None, a=None, a_i=None, p_s=None, p_c=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        mean, z_w = np.dot(self._w, x[order]), np.dot(self._w, z[order])  # Line 3 and 7
        p_c = (1.0-self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*np.dot(a, z_w)  # Line 8
        v = np.dot(a_i, p_c)  # Line 9
        v_norm = np.dot(v, v)  # (||v||)^2
        s_v_norm = np.sqrt(1.0 + self.c_cov/(1.0 - self.c_cov)*v_norm)
        a_i = (a_i - (1.0 - 1.0/s_v_norm)*np.dot(v[:, np.newaxis], np.dot(v[np.newaxis, :], a_i))/v_norm
               )/np.sqrt(1.0 - self.c_cov)  # Line 10
        a = np.sqrt(1.0 - self.c_cov)*(a + (s_v_norm - 1.0)*np.dot(
            p_c[:, np.newaxis], v[np.newaxis, :])/v_norm)  # Line 11
        p_s = (1.0 - self.c_s)*p_s + np.sqrt(self.c_s*(2.0 - self.c_s)*self._mu_eff)*z_w  # Line 12
        self.sigma *= np.exp(self.c_s/self.d_s*(np.linalg.norm(p_s)/self._e_chi - 1.0))  # Line 13
        return mean, a, a_i, p_s, p_c

    def restart_reinitialize(self, mean=None, z=None, x=None, a=None, a_i=None, p_s=None, p_c=None, y=None):
        if ES.restart_reinitialize(self):
            mean, z, x, a, a_i, p_s, p_c, y = self.initialize(True)
        return mean, z, x, a, a_i, p_s, p_c, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, z, x, a, a_i, p_s, p_c, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            z, x, y = self.iterate(z, x, mean, a, y, args)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, a, a_i, p_s, p_c = self._update_distribution(z, x, a, a_i, p_s, p_c, y)
            self._print_verbose_info(y)
            self._n_generations += 1
            if self.is_restart:
                mean, z, x, a, a_i, p_s, p_c, y = self.restart_reinitialize(
                    mean, z, x, a, a_i, p_s, p_c, y)
        return self._collect_results(fitness, mean)



"""Repeat the following paper for `OPOA2010`:
    Arnold, D.V. and Hansen, N., 2010, July.
    Active covariance matrix adaptation for the (1+1)-CMA-ES.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 385-392). ACM.
    https://dl.acm.org/doi/abs/10.1145/1830483.1830556

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/opoa2010

    Luckily our code could repeat the data reported in the original paper *nearly well*.
    Therefore, we argue that the repeatability of `OPOA2010` could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere, ellipsoid, cigar, discus, cigar_discus
from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.opoa2010 import OPOA2010


class Fig1(OPOA2010):
    def optimize(self, fitness_function=None, args=None):
        fitness = ES.optimize(self, fitness_function)
        sigmas = [self.sigma]  # for plotting
        mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args)
        if self.saving_fitness:
            fitness.append(y)
        self._print_verbose_info(y)
        while True:
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.iterate(
                args, mean, a, a_i, best_so_far_y, p_s, p_c)
            sigmas.append(self.sigma)  # for plotting
            if self.saving_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, a, a_i, best_so_far_y, p_s, p_c = self.restart_reinitialize(
                    args, mean, y, a, a_i, best_so_far_y, p_s, p_c, fitness)
        res = self._collect_results(fitness, mean)
        res['sigmas'] = sigmas  # for plotting
        return res


def plot(function, dim):
    res = pickle.load(open(function + '_' + str(dim) + '.pickle', 'rb'))
    plt.figure()
    plt.yscale('log')
    plt.ylim([1e-9, 1e6])
    plt.yticks([1e-9, 1e-6, 1e-3, 1e0, 1e3, 1e6])
    plt.xticks([0, 1e3, 2e3, 3e3, 4e3, 5e3, 6e3])
    plt.xlim([0, 6e3])
    plt.plot(res['fitness'][:, 0], res['fitness'][:, 1],
             color='r', label='function value f(x)')
    plt.plot(np.arange(len(res['sigmas'])), res['sigmas'],
             color='g', label=r'mutation strength $\sigma$')
    plt.xlabel("time t")
    plt.legend()
    plt.show()


def plot2(function, evaluation, dim):
    plt.xticks(rotation=30)
    plt.bar(function, evaluation, color='g', width=0.5)
    if dim == 2:
        plt.ylim([0, 6e2])
        plt.yticks([0, 2e2, 4e2, 6e2])
    elif dim == 4:
        plt.ylim([0, 1500])
        plt.yticks([0, 5e2, 1e3, 1.5e3])
    elif dim == 10:
        plt.ylim([0, 7.2e3])
        plt.yticks([0, 2.4e3, 4.8e3, 7.2e3])
    elif dim == 20:
        plt.ylim([0, 2.4e4])
        plt.yticks([0, 8e3, 1.6e4, 2.4e4])
    elif dim == 40:
        plt.ylim([0, 8.4e4])
        plt.yticks([0, 2.8e4, 5.6e4, 8.4e4])
    plt.title("n = " + str(dim))
    plt.show()


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    d = 10
    problem = {'fitness_function': discus,
               'ndim_problem': d}
    options = {'fitness_threshold': 1e-9,
               'seed_rng': 0,  # not given in the original paper
               'x': np.random.default_rng(1).standard_normal((d,)),  # mean
               'sigma': 0.1,
               'saving_fitness': 1}
    solver = Fig1(problem, options)
    results = solver.optimize()
    pickle.dump(results, open(discus.__name__ + '_' + str(d) + '.pickle', 'wb'))
    plot(discus.__name__, d)

    functions = [sphere, ellipsoid, cigar, discus, cigar_discus]
    for d in [10, 20, 40]:
        evaluations = []
        for f in functions:
            problem = {'fitness_function': f,
                       'ndim_problem': d}
            options = {'fitness_threshold': 1e-10,
                       'seed_rng': 0,
                       'x': np.random.default_rng(1).standard_normal((d,)),  # mean
                       'sigma': 0.1,
                       'is_restart': False,
                       'saving_fitness': 1}
            solver = OPOA2010(problem, options)
            results = solver.optimize()
            evaluations.append(results['n_function_evaluations'])
        plot2([f.__name__ for f in functions], evaluations, d)



import numpy as np

from pypop7.benchmarks import base_functions
from pypop7.benchmarks.base_functions import BaseFunction
from pypop7.benchmarks.shifted_functions import load_shift_vector
from pypop7.benchmarks.rotated_functions import load_rotation_matrix


# helper functions
def load_shift_and_rotation(func, x, shift_vector=None, rotation_matrix=None):
    shift_vector = load_shift_vector(func, x, shift_vector)
    rotation_matrix = load_rotation_matrix(func, x, rotation_matrix)
    return shift_vector, rotation_matrix


def sphere(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(sphere, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.sphere(x)
    return y


class Sphere(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'sphere'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return sphere(x, shift_vector, rotation_matrix)


def cigar(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(cigar, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.cigar(x)
    return y


class Cigar(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return cigar(x, shift_vector, rotation_matrix)


def discus(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(discus, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.discus(x)
    return y


class Discus(BaseFunction):  # also called Tablet
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'discus'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return discus(x, shift_vector, rotation_matrix)


def cigar_discus(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(cigar_discus, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.cigar_discus(x)
    return y


class CigarDiscus(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar_discus'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return cigar_discus(x, shift_vector, rotation_matrix)


def ellipsoid(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(ellipsoid, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.ellipsoid(x)
    return y


class Ellipsoid(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ellipsoid'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return ellipsoid(x, shift_vector, rotation_matrix)


def different_powers(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(different_powers, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.different_powers(x)
    return y


class DifferentPowers(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'different_powers'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return different_powers(x, shift_vector, rotation_matrix)


def schwefel221(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(schwefel221, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.schwefel221(x)
    return y


class Schwefel221(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel221'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return schwefel221(x, shift_vector, rotation_matrix)


def step(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(step, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.step(x)
    return y


class Step(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'step'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return step(x, shift_vector, rotation_matrix)


def schwefel222(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(schwefel222, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.schwefel222(x)
    return y


class Schwefel222(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel222'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return schwefel222(x, shift_vector, rotation_matrix)


def rosenbrock(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(rosenbrock, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.rosenbrock(x)
    return y


class Rosenbrock(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rosenbrock'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return rosenbrock(x, shift_vector, rotation_matrix)


def schwefel12(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(schwefel12, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.schwefel12(x)
    return y


class Schwefel12(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel12'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return schwefel12(x, shift_vector, rotation_matrix)


def exponential(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(exponential, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.exponential(x)
    return y


class Exponential(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'exponential'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return exponential(x, shift_vector, rotation_matrix)


def griewank(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(griewank, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.griewank(x)
    return y


class Griewank(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'griewank'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return griewank(x, shift_vector, rotation_matrix)


def bohachevsky(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(bohachevsky, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.bohachevsky(x)
    return y


class Bohachevsky(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'bohachevsky'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return bohachevsky(x, shift_vector, rotation_matrix)


def ackley(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(ackley, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.ackley(x)
    return y


class Ackley(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ackley'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return ackley(x, shift_vector, rotation_matrix)


def rastrigin(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(rastrigin, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.rastrigin(x)
    return y


class Rastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rastrigin'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return rastrigin(x, shift_vector, rotation_matrix)


def scaled_rastrigin(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(scaled_rastrigin, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.scaled_rastrigin(x)
    return y


class ScaledRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'scaled_rastrigin'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return scaled_rastrigin(x, shift_vector, rotation_matrix)


def skew_rastrigin(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(skew_rastrigin, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.skew_rastrigin(x)
    return y


class SkewRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'skew_rastrigin'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return skew_rastrigin(x, shift_vector, rotation_matrix)


def levy_montalvo(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(levy_montalvo, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.levy_montalvo(x)
    return y


class LevyMontalvo(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'levy_montalvo'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return levy_montalvo(x, shift_vector, rotation_matrix)


def michalewicz(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(michalewicz, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.michalewicz(x)
    return y


class Michalewicz(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'michalewicz'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return michalewicz(x, shift_vector, rotation_matrix)


def salomon(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(salomon, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.salomon(x)
    return y


class Salomon(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'salomon'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return salomon(x, shift_vector, rotation_matrix)


def shubert(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(shubert, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.shubert(x)
    return y


class Shubert(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'shubert'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return shubert(x, shift_vector, rotation_matrix)


def schaffer(x, shift_vector=None, rotation_matrix=None):
    shift_vector, rotation_matrix = load_shift_and_rotation(schaffer, x, shift_vector, rotation_matrix)
    x = np.dot(rotation_matrix, x - shift_vector)
    y = base_functions.schaffer(x)
    return y


class Schaffer(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schaffer'

    def __call__(self, x, shift_vector=None, rotation_matrix=None):
        return schaffer(x, shift_vector, rotation_matrix)



"""Repeat the following paper for `GL25`:
    García-Martínez, C., Lozano, M., Herrera, F., Molina, D. and Sánchez, A.M., 2008.
    Global and local real-coded genetic algorithms based on parent-centric crossover operators.
    European Journal of Operational Research, 185(3), pp.1088-1113.
    https://www.sciencedirect.com/science/article/abs/pii/S0377221706006308

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `GL25` could be **well-documented**.
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, schwefel222
from pypop7.optimizers.ga.gl25 import GL25 as Solver


if __name__ == '__main__':
    ndim_problem = 25

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'initial_lower_boundary': 4 * np.ones((ndim_problem,)),
               'initial_upper_boundary': 5 * np.ones((ndim_problem,)),
               'lower_boundary': -5.12 * np.ones((ndim_problem,)),
               'upper_boundary': 5.12 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 100000,
               'seed_rng': 0,
               'saving_fitness': 1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 3.683380983946287e-151 vs 4.36e-147 (from the original paper)

    problem = {'fitness_function': schwefel222,
               'ndim_problem': ndim_problem,
               'initial_lower_boundary': 8 * np.ones((ndim_problem,)),
               'initial_upper_boundary': 10 * np.ones((ndim_problem,)),
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 100000,
               'seed_rng': 0,
               'saving_fitness': 1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 2.8901400348477053e-79 vs 9.85e-77 (from the original paper)



import numpy as np

from pypop7.optimizers.es.es import ES


class FCMAES(ES):
    """Fast Covariance Matrix Adaptation Evolution Strategies(FCMAES, (μ/μ_w,λ)-Fast CMA-ES)
        Reference
        ---------------
        Z. Li, Q. Zhang, X. Lin, H. zhen
        Fast Covariance Matrix Adaptation for Large-Scale Black-Box Optimization
        IEEE Transaction on Cybernetics vol.50 No.5 May 2020
        https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=8533604
    """
    def __init__(self, prolblem, options):
        ES.__init__(self, prolblem, options)
        self.n_evolution_paths = self.n_individuals
        self.c = 2 / (self.ndim_problem + 5)
        self.c_1 = 1 / (3 * np.sqrt(self.ndim_problem) + 5)
        self.c_sigma = 0.3
        self.d_sigma = 1
        self.q_star = 0.27
        self.T = self.ndim_problem
        self.sigma_threshold = 1e-12
        self.fitness_diff = 1e-12
        self.fitness_rel_diff = 1e-6
        self.s_1 = 1 - self.c_1
        self.s_2 = np.sqrt((1 - self.c_1) * self.c_1)
        self.s_3 = np.sqrt(self.c_1)
        self.d_1 = 1 - self.c
        self.d_2 = np.sqrt(self.c * (2 - self.c) * self._mu_eff)

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals, ))
        p = np.zeros((self.ndim_problem, ))
        p_hat = np.zeros((self.n_evolution_paths, self.ndim_problem))
        s = 0
        f = np.inf * np.ones((self.n_parents, ))
        return mean, x, y, p, p_hat, s, f

    def iterate(self, mean, x, y, p, p_hat):
        for i in range(self.n_individuals):
            z = self.rng_optimization.standard_normal((self.ndim_problem, ))
            r1 = self.rng_optimization.standard_normal()
            r2 = self.rng_optimization.standard_normal()
            x[i] = mean + self.sigma * (self.s_1 * z + self.s_2 * r1 * p_hat[i] + self.s_3 * r2 * p)
            y[i] = self._evaluate_fitness(x[i])
        return x, y

    def _update_distribution(self, mean, x, y, p, p_hat, s, f):
        new_mean = 0
        new_f = np.empty((self.n_parents, ))
        order = np.argsort(y)
        r1, r2 = np.empty((self.n_parents, )), np.empty((self.n_parents, ))
        for i in range(self.n_parents):
            new_mean += self._w[i] * x[order[i]]
            new_f[i] = y[order[i]]
        p = self.d_1 * p + self.d_2 * (new_mean - mean) / self.sigma
        mean = new_mean
        if self._n_generations % self.T == 0:
            for i in range(self.n_evolution_paths - 1):
                p_hat[i] = p_hat[i+1]
            p_hat[-1] = p
        F = np.append(f, new_f, axis=0)
        F = np.sort(F)
        p1, p2 = 0, 0
        for i in range(len(F)):
            if p1 != self.n_parents and f[p1] == F[i]:
                r1[p1] = i
                p1 += 1
            elif new_f[p2] == F[i]:
                r2[p2] = i
                p2 += 1
        assert p1 == self.n_parents
        assert p2 == self.n_parents
        q = 0
        for i in range(self.n_parents):
            q += self._w[i] * (r1[i] - r2[i])
        q /= self.n_parents
        f = new_f
        s = (1 - self.c_sigma) * s + self.c_sigma * (q - self.q_star)
        self.sigma *= np.exp(s / self.d_sigma)
        return mean, p, p_hat, s, f

    def restart_initialize(self, args=None, mean=None, x=None, y=None, p=None, p_hat=None,
                           s=None, f=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2, is_restart_3 = self.sigma < self.sigma_threshold, False, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
            is_restart_3 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) /\
                           self._fitness_list[-1] < self.fitness_rel_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2) or bool(is_restart_3)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            self.n_individuals *= 2
            self.n_parents = int(self.n_individuals / 2)
            self.n_evolution_paths = self.n_individuals
            w_base, w = np.log((self.n_individuals + 1) / 2), np.log(np.arange(self.n_parents) + 1)
            self._w = (w_base - w) / (self.n_parents * w_base - np.sum(w))
            self._mu_eff = 1 / np.sum(np.power(self._w, 2))
            self._fitness_list = [np.Inf]
            self.d_sigma *= 2
            mean, x, y, p, p_hat, s, f = self.initialize(args, is_restart)
        return mean, x, y, p, p_hat, s, f

    def optimize(self, fitness_function=None, args=None):
        fitness = ES.optimize(self, fitness_function)
        mean, x, y, p, p_hat, s, f = self.initialize()
        while True:
            x, y = self.iterate(mean, x, y, p, p_hat)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, p_hat, s, f = self._update_distribution(mean, x, y, p, p_hat, s, f)
            self._n_generations += 1
            self._print_verbose_info(y)
            mean, x, y, p, p_hat, s, f = self.restart_initialize(args, mean, x, y, p, p_hat, s, f)
        results = self._collect_results(fitness, mean)
        return results



"""Repeat the following paper for `ARHC`:
    Schaul, T., Bayer, J., Wierstra, D., Sun, Y., Felder, M., Sehnke, F., Rückstieß, T. and Schmidhuber, J., 2010.
    PyBrain.
    Journal of Machine Learning Research, 11(24), pp.743-746.
    https://jmlr.org/papers/v11/schaul10a.html

    Luckily our code could repeat the data reported in the other code *well*.
    Therefore, we argue that the repeatability of `ARHC` could be **well-documented**.

    The baseline Python code is shown below:
    ----------------------------------------
    import time

    import numpy as np
    # use Python 2.7 to avoid possible unsuccessful installation, since pybrain is not maintained now
    from pybrain.optimization.hillclimber import StochasticHillClimber as ARHC


    def ellipsoid(x):
        x = np.power(x, 2)
        y = np.dot(np.power(10, 6 * np.linspace(0, 1, x.size)), x)
        return y


    solver = ARHC(ellipsoid, 4 * np.ones((1000,)), minimize=True, maxEvaluations=2e6, verbose=True)
    solver.temperature = 100
    start_time = time.time()
    solver.learn()
    # Step: 1999998 best: 10972437.61259182 (different runs result in slightly different results)
    print("Runtime: {:7.5e}".format(time.time() - start_time))
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import ellipsoid
from pypop7.optimizers.rs.arhc import ARHC


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 1000
    problem = {'fitness_function': ellipsoid,
               'ndim_problem': ndim_problem,
               'upper_boundary': 5.0 * np.ones((ndim_problem,)),
               'lower_boundary': -5.0 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2e6,
               'seed_rng': 0,
               'x': 4 * np.ones((ndim_problem,)),
               'sigma': 0.1,
               'temperature': 100,
               'verbose': 200000,
               'saving_fitness': 200000}
    arhc = ARHC(problem, options)
    results = arhc.optimize()
    print(results)  # 1.03891664e+07
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



"""PyPop7 is a Pure-PYthon library of POPulation-based OPtimization for single-objective, real-parameter, black-box
    problems (currently still actively developed). Its main goal is to provide a unified interface and elegant
    implementations for Black-Box Optimization (BBO), particularly population-based optimizers, in order to
    facilitate research repeatability and also real-world applications.

    More specifically, for alleviating the notorious curse of dimensionality of BBO (almost based on iterative
    sampling), the primary focus of PyPop7 is to cover their State-Of-The-Art (SOTA) implementations for
    Large-Scale Optimization (LSO), though many of their other versions and variants are also included here
    (for benchmarking/mixing purpose, and sometimes even for practical purpose).
"""
from pypop7.optimizers.core import Optimizer, Terminations
from pypop7.optimizers.rs import RS, PRS, RHC, ARHC, SRS, BES
from pypop7.optimizers.ga import GA, GENITOR, G3PCX, GL25, ASGA
from pypop7.optimizers.sa import SA, CSA, ESA, NSA
from pypop7.optimizers.pso import PSO, SPSO, SPSOL, CLPSO, IPSO, CCPSO2



import numpy as np

from pypop7.optimizers.rs.rhc import RHC


class ARHC(RHC):
    """Annealed Random Hill Climber (ARHC).

    .. note:: The search performance of `ARHC` depends **heavily** on the *temperature* setting of the annealing
       process. However, its setting is a **non-trival** task, since it may vary among different problems and even
       among different optimization stages for the problem at hand. Therefore, it is **highly recommended** to
       first attempt other more advanced (e.g. population-based) methods for large-scale black-box optimization.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'x'                           - initial (starting) point (`array_like`),
                * 'sigma'                       - initial global step-size (`float`),
                * 'temperature'                 - annealing temperature (`float`),
                * 'initialization_distribution' - random sampling distribution for starting point initialization (`int`,
                  default: `1`). Only when `x` is not set *explicitly*, it will be used.

                  * `1`: *uniform* distribution is used for random sampling,
                  * `0`: *standard normal* distribution is used for random sampling with mean `0` and std `1` for
                    each dimension.

    Examples
    --------
    Use the Random Search optimizer `ARHC` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.rs.arhc import ARHC
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1,
       ...            'temperature': 1.5}
       >>> arhc = ARHC(problem, options)  # initialize the optimizer class
       >>> results = arhc.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"ARHC: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       ARHC: 5000, 0.0002641143073543329

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/2s3v8z7h>`_ for more details.

    Attributes
    ----------
    initialization_distribution : `int`
                                  random sampling distribution for initialization of starting point.
    sigma                       : `float`
                                  initial global step-size.
    temperature                 : `float`
                                  annealing temperature.
    x                           : `array_like`
                                  initial (starting) point.

    References
    ----------
    Russell, S. and Norvig P., 2021.
    Artificial intelligence: A modern approach (Global Edition).
    Pearson Education.
    http://aima.cs.berkeley.edu/    (See CHAPTER 4: SEARCH IN COMPLEX ENVIRONMENTS)

    https://github.com/pybrain/pybrain/blob/master/pybrain/optimization/hillclimber.py
    """
    def __init__(self, problem, options):
        RHC.__init__(self, problem, options)
        self.temperature = options.get('temperature')  # annealing temperature
        if self.temperature is None:
            raise ValueError('`temperature` should be set.')
        self._parent_x = np.copy(self.best_so_far_x)
        self._parent_y = np.copy(self.best_so_far_y)

    def iterate(self):  # sampling via mutating the parent individual
        noise = self.rng_optimization.standard_normal(size=(self.ndim_problem,))
        return self._parent_x + self.sigma*noise

    def _evaluate_fitness(self, x, args=None):
        y = RHC._evaluate_fitness(self, x, args)
        # update parent solution and fitness
        diff = y - self._parent_y
        if (diff < 0) or (self.rng_optimization.random() < np.exp(-diff/self.temperature)):
            self._parent_x, self._parent_y = np.copy(x), y
        return y



"""Repeat the following paper for `CCMAES2009`:
    Krause, O. and Igel, C., 2015, January.
    A more efficient rank-one covariance matrix update for evolution strategies.
    In Proceedings of ACM Conference on Foundations of Genetic Algorithms (pp. 129-136).
    https://dl.acm.org/doi/abs/10.1145/2725494.2725496

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/ccmaes2009

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `CCMAES2009` could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.optimizers.es.ccmaes2009 import CCMAES2009 as Solver
from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock


def plot(function, ndim):
    plt.figure()
    res = pickle.load(open(function + '_' + str(ndim) + '.pickle', 'rb'))
    plt.yscale('log')
    if function == 'ellipsoid':
        plt.ylim([1e-35, 1e5])
        plt.yticks([1e-35, 1e-30, 1e-25, 1e-20, 1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 3e4])
        plt.xticks([0, 5e3, 1e4, 1.5e4, 2e4, 2.5e4, 3e4])
    elif function == 'rosenbrock':
        plt.ylim([1e-20, 1e5])
        plt.yticks([1e-20, 1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 3e4])
        plt.xticks([0, 5e3, 1e4, 1.5e4, 2e4, 2.5e4, 3e4])
    plt.plot(res['fitness'][:, 0], res['fitness'][:, 1], color='r')
    plt.xlabel("objective function evaluations")
    plt.ylabel("objective function value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    for f in [ellipsoid, rosenbrock]:
        for d in [20]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': 0.1 * np.ones((d,)),
                       'upper_boundary': 0.3 * np.ones((d,))}
            options = {'fitness_threshold': 1e-10,
                       'seed_rng': 0,
                       'sigma': 0.2 / 3,
                       'saving_fitness': 1,
                       'is_restart': False}
            if f.__name__ == 'ellipsoid':
                options['fitness_threshold'] = 1e-35
                options['max_function_evaluations'] = 3e4
            elif f.__name__ == 'rosenbrock':
                options['fitness_threshold'] = 1e-20
                options['max_function_evaluations'] = 3e4
            solver = Solver(problem, options)
            results = solver.optimize()
            pickle.dump(results, open(f.__name__ + '_' + str(d) + '.pickle', 'wb'))
            plot(f.__name__, d)



import numpy as np

from pypop7.optimizers.pso.pso import PSO


class SPSOL(PSO):
    """Standard Particle Swarm Optimizer with a Local (ring) topology (SPSOL).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, aka number of particles (`int`, default: `20`),
                * 'cognition'     - cognitive learning rate (`float`, default: `2.0`),
                * 'society'       - social learning rate (`float`, default: `2.0`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Examples
    --------
    Use the optimizer `SPSOL` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.pso.spsol import SPSOL
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> spsol = SPSOL(problem, options)  # initialize the optimizer class
       >>> results = spsol.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SPSOL: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SPSOL: 5000, 3.470837498146212e-08

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mucj6b6y>`_ for more details.

    Attributes
    ----------
    cognition     : `float`
                    cognitive learning rate, aka acceleration coefficient.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.
    n_individuals : `int`
                    swarm (population) size, aka number of particles.
    society       : `float`
                    social learning rate, aka acceleration coefficient.

    References
    ----------
    Blackwell, T. and Kennedy, J., 2018.
    Impact of communication topology in particle swarm optimization.
    IEEE Transactions on Evolutionary Computation, 23(4), pp.689-702.
    https://ieeexplore.ieee.org/abstract/document/8531770

    Floreano, D. and Mattiussi, C., 2008.
    Bio-inspired artificial intelligence: Theories, methods, and technologies.
    MIT Press.
    https://mitpress.mit.edu/9780262062718/bio-inspired-artificial-intelligence/
    (See [Chapter 7.2 Particle Swarm Optimization] for details.)

    Venter, G. and Sobieszczanski-Sobieski, J., 2003.
    Particle swarm optimization.
    AIAA Journal, 41(8), pp.1583-1589.
    https://arc.aiaa.org/doi/abs/10.2514/2.2111

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146

    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Eberhart, R. and Kennedy, J., 1995, October.
    A new optimizer using particle swarm theory.
    In Proceedings of International Symposium on Micro Machine and Human Science (pp. 39-43). IEEE.
    https://ieeexplore.ieee.org/abstract/document/494215
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)
        assert self.n_individuals >= 3  # for ring topology

    def _ring_topology(self, p_x=None, p_y=None, i=None):
        left, right = i - 1, i + 1
        if i == 0:
            left = self.n_individuals - 1
        elif i == self.n_individuals - 1:
            right = 0
        ring = [left, i, right]
        return p_x[ring[int(np.argmin(p_y[ring]))]]

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            n_x[i] = self._ring_topology(p_x, p_y, i)  # online update within ring topology
            cognition_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            society_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            v[i] = (self._w[self._n_generations]*v[i] +
                    self.cognition*cognition_rand*(p_x[i] - x[i]) +
                    self.society*society_rand*(n_x[i] - x[i]))  # velocity update
            v[i] = np.clip(v[i], self._min_v, self._max_v)
            x[i] += v[i]  # position update
            y[i] = self._evaluate_fitness(x[i], args)  # fitness evaluation
            if y[i] < p_y[i]:  # online update
                p_x[i], p_y[i] = x[i], y[i]
        self._n_generations += 1
        return v, x, y, p_x, p_y, n_x



"""Repeat the following paper for `CLPSO`:
    Liang, J.J., Qin, A.K., Suganthan, P.N. and Baskar, S., 2006.
    Comprehensive learning particle swarm optimizer for global optimization of multimodal functions.
    IEEE Transactions on Evolutionary Computation, 10(3), pp.281-295.
    https://ieeexplore.ieee.org/abstract/document/1637688

    Luckily our code could repeat the data generated by the original paper *well*.
    Therefore, we argue that the repeatability of `CLPSO` could be **well-documented**.
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import sphere, ackley, rastrigin
from pypop7.optimizers.pso.clpso import CLPSO as Solver


if __name__ == '__main__':
    ndim_problem = 30

    start_run = time.time()
    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,)),
               'initial_lower_boundary': -100 * np.ones((ndim_problem,)),
               'initial_upper_boundary': 50 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 200000,
               'seed_rng': 0,  # undefined in the original paper
               'n_individuals': 40,
               'verbose': 1e3,
               'saving_fitness': 20000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 6.10600207e-17 vs 4.46e-14 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))

    start_run = time.time()
    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32.768 * np.ones((ndim_problem,)),
               'upper_boundary': 32.768 * np.ones((ndim_problem,)),
               'initial_lower_boundary': -32.768 * np.ones((ndim_problem,)),
               'initial_upper_boundary': 16 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 200000,
               'seed_rng': 0,  # undefined in the original paper
               'n_individuals': 40,
               'm': 5,
               'verbose': 1e3,
               'saving_fitness': 20000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 9.16746679e-11 vs 0 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))

    start_run = time.time()
    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12 * np.ones((ndim_problem,)),
               'upper_boundary': 5.12 * np.ones((ndim_problem,)),
               'initial_lower_boundary': -5.12 * np.ones((ndim_problem,)),
               'initial_upper_boundary': 2 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 200000,
               'seed_rng': 0,  # undefined in the original paper
               'n_individuals': 40,
               'verbose': 1e3,
               'saving_fitness': 20000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 3.99211331e-10 vs 4.85e-10 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np

from pypop7.optimizers.cc.cc import CC


class COSYNE(CC):
    """Cooperative Synapse Neuroevolution(CoSyNE)

    Parameters
    ----------
     problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
              and with the following particular settings (`keys`):
                * 'n_individuals'            - number of individuals/samples (`int`),
                * 'n_combine'                - number of combine (`int`, default: `2`),
                * 'crossover_type'           - type of crossover operation (`string`, default: `one_point`),
                * 'prob_mutate'              - probability to mutate (`float`, default: `0.3`).

    Examples
    --------
    Use the Cooperative Evolution optimizer `COSYNE` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:
    .. code-block:: python
       :linenos:
       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import ellipsoid # function to be minimized
       >>> from pypop7.optimizers.cc.cosyne import COSYNE
       >>> problem = {'fitness_function': ellipsoid,  # define problem arguments
       ...            'ndim_problem': 10,
       ...            'lower_boundary': -5 * numpy.ones((10,)),
       ...            'upper_boundary': 5 * numpy.ones((10,))}
       >>> options = {'max_function_evaluations': 3e5,  # set optimizer options
       ...            'n_individuals': 20,
       ...            'n_combine': 2,
       ...            'prob_mutate': 0.3,
       ...            'crossover_type': "one_point"}
       >>> cosyne = COSYNE(problem, options)  # initialize the optimizer class
       >>> results = cosyne.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"COSYNE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       COSYNE: 300061, 0.561405461371189(1.0746113061904907 in EvoTorch)

    Attributes
    ----------
    prob_mutate      : `float`
                        probability to mutate.
    crossover_type   : `string`
                        type of crossover.
    n_combine        : `int`
                        number of combine individuals.
    n_individuals    : `int`
                        number of individuals/samples.

    Reference
    ---------
    F. Gomez, J. Schmidhuber, R. Miikkulainen
    Accelerated Neural Evolution through Cooperatively Coevolved Synapses
    https://jmlr.org/papers/v9/gomez08a.html
    """
    def __init__(self, problem, options):
        CC.__init__(self, problem, options)
        self.n_combine = options.get('n_combine', 2)
        self.crossover_type = options.get('crossover_type', "one_point")
        self.prob_mutate = options.get('prob_mutate', 0.3)

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            x[i] = self._initialize_x()
        return x, y

    def mutate(self, x):
        for i in range(2):
            for j in range(self.ndim_problem):
                rand = np.random.random()
                if rand < self.prob_mutate:
                    x[i][j] = x[i][j] + 0.3 * self.rng_optimization.standard_cauchy(1)
            x[i] = np.clip(x[i], self.lower_boundary, self.upper_boundary)
        return x

    def permute(self, log):
        temp = []
        length = len(log)
        for k in range(length):
            index = np.random.randint(0, len(log)) % len(log)
            temp.append(log[index])
            log.pop(index)
        return temp

    def iterate(self, x, y):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            y[i] = self._evaluate_fitness(x[i])
        order = np.argsort(y)
        o = self.crossover(x[order[0]], x[order[1]], self.crossover_type)
        o = self.mutate(o)
        for i in range(self.ndim_problem):
            order_1 = np.argsort(y)
            for k in range(2):
                x[order_1[self.n_individuals - k - 1]][i] = o[k][i]
            mark_weight, log = [], []
            for j in range(self.n_individuals):
                temp = (y[j] - y[order[0]])/(y[order[-1]] - y[order[0]])
                prob = 1 - np.power(temp, 1.0 / self.ndim_problem)
                rand = np.random.random()
                if rand < prob:
                    mark_weight.append(x[j][i])
                    log.append(j)
            log = self.permute(log)
            for j in range(len(log)):
                x[log[j]][i] = mark_weight[j]
                y[log[j]] = self._evaluate_fitness(x[log[j]])
        return x, y

    def optimize(self, fitness_function=None):
        fitness = CC.optimize(self, fitness_function)
        x, y = self.initialize()
        while True:
            x, y = self.iterate(x, y)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



"""Repeat the following paper for `CCMAES2016`:
    Krause, O., Arbonès, D.R. and Igel, C., 2016.
    CMA-ES with optimal covariance update and storage complexity.
    Advances in Neural Information Processing Systems, 29, pp.370-378.
    https://proceedings.neurips.cc/paper/2016/hash/289dff07669d7a23de0ef88d2f7129e7-Abstract.html

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/ccmaes2016
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.es.ccmaes2016 import CCMAES2016


def cigar(x):
    x = np.power(x, 2)
    y = 1e-6*x[0] + np.sum(x[1:])
    return y


def discus(x):  # also called tablet
    x = np.power(x, 2)
    y = x[0] + 1e-6*np.sum(x[1:])
    return y


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    dims = [4, 8, 16, 32, 64, 128, 256]
    for f in [sphere, cigar, discus]:
        plt.figure()
        plt.ylim([1e2, 1e4])
        plt.xscale('log')
        plt.yticks([1e2, 1e3, 1e4])
        plt.ylabel('iterations')
        plt.yscale('log')
        n_iterations = []
        for d in [4, 8, 16, 32, 64, 128, 256]:
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': np.zeros((d,)),
                       'upper_boundary': np.ones((d,))}
            options = {'seed_rng': 0,  # not given in the original paper
                       'fitness_threshold': 1e-14,
                       'sigma': 1 / 3,  # not given in the original paper
                       'saving_fitness': 1,
                       'is_restart': False}
            if f == sphere:
                options['x'] = np.random.default_rng(1).standard_normal((d,))  # mean
            solver = CCMAES2016(problem, options)
            results = solver.optimize()
            n_iterations.append(results['_n_generations'])
        print(n_iterations)
        plt.plot(dims, n_iterations, color='r')
        plt.semilogx(base=2)
        plt.xlim([4, 256])
        plt.xticks(dims)
        plt.show()



import numpy as np

from pypop7.optimizers.ga.ga import GA
from athena.active import ActiveSubspaces


class ASGA(GA):
    """Active Subspace Genetic Algorithm (ASGA).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_initial_individuals' - initial population size (`int`, default: `2000`),
                * 'n_individuals'         - population size (`int`, default: `200`),
                * 'n_subspace'            - dimensionality number of active subspaces (`int`, default: `1`),
                * 'crossover_prob'        - crossover probability (`float`, default: `0.5`),
                * 'mutation_prob'         - mutation probability (`float`, default: `0.5`),
                * 'b'                     - number of back-mapped points (`int`, default: `2`).

    Examples
    --------
    Use the Genetic Algorithm optimizer `ASGA` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ga.asga import ASGA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> asga = ASGA(problem, options)  # initialize the optimizer class
       >>> results = asga.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"ASGA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       ASGA: 5000, 0.0072316975092846245

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3z8zxr74>`_ for more details.

    Attributes
    ----------
    b                     : `int`
                            number of back-mapped points.
    crossover_prob        : `float`
                            crossover probability.
    mutation_prob         : `float`
                            mutation probability.
    n_individuals         : `int`
                            population size.
    n_initial_individuals : `int`
                            initial population size.
    n_subspace            : `int`
                            dimensionality number of active subspaces.

    References
    ----------
    Demo, N., Tezzele, M. and Rozza, G., 2021.
    A supervised learning approach involving active subspaces for an efficient genetic algorithm in
    high-dimensional optimization problems.
    SIAM Journal on Scientific Computing, 43(3), pp.B831-B853.
    https://epubs.siam.org/doi/10.1137/20M1345219
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.crossover_prob = options.get('crossover_prob', 0.5)  # crossover probability
        self.mutation_prob = options.get('mutation_prob', 0.5)  # mutation probability
        self.n_subspace = options.get('n_subspace', 1)  # dimensionality number of active subspaces
        self.b = options.get('b', 2)  # number of back-mapped points
        self.alpha = options.get('alpha', 1.0)
        self.n_initial_individuals = options.get('n_initial_individuals', 2000)
        self.n_individuals = options.get('n_individuals', 200)
        self._n_individuals_subspace = int(self.n_individuals/self.b)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_initial_individuals, self.ndim_problem))  # initial population
        y = np.empty((self.n_initial_individuals,))  # fitness
        for i in range(self.n_initial_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        x_as, y_as = np.copy(x), np.copy(y)
        return x, y, x_as, y_as

    def _build_active_space(self, x_as=None, y_as=None):
        active_subspace = ActiveSubspaces(dim=self.n_subspace, method='local')
        active_subspace.fit(inputs=x_as, outputs=y_as)
        return active_subspace

    def _select(self, x=None, y=None):
        return x[np.argsort(y)[:self._n_individuals_subspace]]

    def _crossover(self, x=None):  # different from blend BLX-alpha crossover
        xx = np.copy(x)
        for i in range(self._n_individuals_subspace):
            x1, x2 = self.rng_optimization.choice(x, 2)
            for j in range(self.n_subspace):
                if self.rng_optimization.random() < self.crossover_prob:
                    r = self.rng_optimization.uniform(-1.0*self.alpha, 1.0+self.alpha)
                    xx[i][j] = (1.0 - r)*x1[j] + r*x2[j]
        return xx

    def _mutate(self, x=None):
        for i in range(self._n_individuals_subspace):
            for j in range(self.n_subspace):
                if self.rng_optimization.random() < self.mutation_prob:
                    x[i][j] *= (1 + self.rng_optimization.normal(0, 0.1))
        return x

    def iterate(self, x=None, y=None, x_as=None, y_as=None, args=None):
        active_subspace = self._build_active_space(x_as, y_as)
        xx = self._select(x, y)
        xx = active_subspace.transform(xx)[0]  # forward reduction
        xx = self._mutate(self._crossover(xx))
        x = active_subspace.inverse_transform(xx, self.b)[0]  # backward mapping
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        x_as, y_as = np.vstack((x_as, x)), np.hstack((y_as, y))
        return x, y, x_as, y_as

    def optimize(self, fitness_function=None, args=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, x_as, y_as = self.initialize(args)
        if self.saving_fitness:
            fitness.extend(y)
        self._print_verbose_info(y)
        while not self._check_terminations():
            try:
                x, y, x_as, y_as = self.iterate(x, y, x_as, y_as, args)
                if self.saving_fitness:
                    fitness.extend(y)
                if self._check_terminations():
                    break
                self._n_generations += 1
                self._print_verbose_info(y)
            except np.linalg.LinAlgError:
                x, y, x_as, y_as = self.initialize(args)
                if self.saving_fitness:
                    fitness.extend(y)
                self._n_generations = 0
                self._print_verbose_info(y)
            except ValueError:
                x, y, x_as, y_as = self.initialize(args)
                if self.saving_fitness:
                    fitness.extend(y)
                self._n_generations = 0
                self._print_verbose_info(y)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.de.de import DE
from pypop7.optimizers.de.cde import CDE


class CODE(CDE):
    """COmposite Differential Evolution (CODE).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_individuals' - population size (`int`, default: `100`).

    Examples
    --------
    Use the Differential Evolution optimizer `CODE` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.de.code import CODE
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 0}
       >>> code = CODE(problem, options)  # initialize the optimizer class
       >>> results = code.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CODE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CODE: 5000, 0.01052980838183792

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, aka offspring population size.

    References
    ----------
    Wang, Y., Cai, Z., and Zhang, Q. 2011.
    Differential evolution with composite trial vector generation strategies and control parameters.
    IEEE Transactions on Evolutionary Computation, 15(1), pp.55–66.
    https://doi.org/10.1109/TEVC.2010.2087271
    """
    def __init__(self, problem, options):
        CDE.__init__(self, problem, options)
        self.boundary = options.get('boundary', False)
        self._pool = [[1.0, 0.1], [1.0, 0.9], [0.8, 0.2]]  # a pool of two control parameters (f, cr)

    def bound(self, x=None):
        if not self.boundary:
            return x
        for k in range(self.n_individuals):
            idx = np.array(x[k] < self.lower_boundary)
            if idx.any():
                x[k][idx] = np.minimum(self.upper_boundary, 2.0*self.lower_boundary - x[k])[idx]
            idx = np.array(x[k] > self.upper_boundary)
            if idx.any():
                x[k][idx] = np.maximum(self.lower_boundary, 2.0*self.upper_boundary - x[k])[idx]
        return x

    def mutate(self, x=None):
        x1 = np.empty((self.n_individuals, self.ndim_problem))
        x2 = np.empty((self.n_individuals, self.ndim_problem))
        x3 = np.empty((self.n_individuals, self.ndim_problem))
        # randomly select from the parameter candidate pool
        base = np.arange(self.n_individuals)
        f_p = self.rng_optimization.choice(self._pool, (self.n_individuals, 3))
        for k in range(self.n_individuals):
            base_k = np.setdiff1d(base, k)
            r = self.rng_optimization.choice(base_k, (3,), False)
            x1[k] = x[r[0]] + f_p[k, 0, 0]*(x[r[1]] - x[r[2]])  # rand/1/bin
            r = self.rng_optimization.choice(base_k, (5,), False)
            x2[k] = (x[r[0]] + self.rng_optimization.random()*(x[r[1]] - x[r[2]]) +
                     f_p[k, 1, 0]*(x[r[3]] - x[r[4]]))  # rand/2/bin
            r = self.rng_optimization.choice(base_k, (3,), False)
            x3[k] = (x[k] + self.rng_optimization.random()*(x[r[0]] - x[k]) +
                     f_p[k, 2, 0]*(x[r[1]] - x[r[2]]))  # current-to-rand/1
        return x1, x2, x3, f_p

    def crossover(self, x_mu=None, x=None, p_cr=None):
        x_cr = np.copy(x)
        for k in range(self.n_individuals):
            j_rand = self.rng_optimization.integers(self.ndim_problem)
            for i in range(self.ndim_problem):
                if (i == j_rand) or (self.rng_optimization.random() < p_cr[k]):
                    x_cr[k, i] = x_mu[k, i]
        return x_cr

    def select(self, args=None, x=None, y=None, x_cr=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                break
            yy = self._evaluate_fitness(x_cr[k], args)
            if yy < y[k]:
                x[k], y[k] = x_cr[k], yy
        return x, y

    def iterate(self, args=None, x=None, y=None):
        x1, x2, x3, f_p = self.mutate(x)
        x1 = self.bound(self.crossover(x1, x, f_p[:, 0, 1]))
        x2 = self.bound(self.crossover(x2, x, f_p[:, 1, 1]))
        x3 = self.bound(x3)
        x, y = self.select(args, x, y, x1)
        x, y = self.select(args, x, y, x2)
        x, y = self.select(args, x, y, x3)
        self._n_generations += 1
        return x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = DE.optimize(self, fitness_function)
        x, y = self.initialize(args)
        if self.saving_fitness:
            fitness.extend(y)
        self._print_verbose_info(y)
        while True:
            x, y = self.iterate(args, x, y)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._print_verbose_info(y)
        return self._collect_results(fitness)



"""Repeat the following paper for `R1ES`:
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/r1es

    Luckily our code could repeat the data reported in the original paper *well*.
    Therefore, we argue that the repeatability of `R1ES` could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import cigar, ellipsoid, discus, rosenbrock
from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.r1es import R1ES


def plot(function, ndim):
    plt.figure()
    result = pickle.load(open(function + '_' + str(ndim) + '.pickle', 'rb'))
    plt.yscale('log')
    plt.xscale('log')
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='purple')
    plt.xlim([1e2, 1e8])
    plt.xticks([1e2, 1e4, 1e6, 1e8])
    if function == 'ellipsoid':
        plt.ylim([1e-8, 1e10])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    elif function == 'discus':
        plt.ylim([1e-8, 1e6])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4])
    elif function == 'rosenbrock':
        plt.ylim([1e-8, 1e10])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    plt.xlabel("Function Evaluations")
    plt.ylabel("Objective Value")
    plt.title(function.capitalize())
    plt.show()


class Fig2(R1ES):  # to save data for Fig. 2
    def optimize(self, fitness_function=None, args=None):
        fit = ES.optimize(self, fitness_function)
        xx, mean, p, s, y = self.initialize(args)
        fit.append(y[0])
        e1 = np.hstack((1, np.zeros((self.ndim_problem - 1,))))  # for similarity between p and e1
        f, p_norm, stepsize, theta = [], [], [], []  # for plotting data
        while True:
            y_bak = np.copy(y)
            xx, y = self.iterate(xx, mean, p, y, args)
            if self.saving_fitness:
                fit.extend(y)
            if self._check_terminations():
                break
            mean, p, s = self._update_distribution(xx, mean, p, s, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            xx, mean, p, s, y = self.restart_reinitialize(args, xx, mean, p, s, y, fit)
            f.append(np.min(y))
            p_norm.append(np.linalg.norm(p))
            stepsize.append(self.sigma)
            theta.append(np.arccos(np.abs(np.dot(p, e1)) / p_norm[-1]))
        res = self._collect_results(fit, mean)
        res['p'] = p
        res['s'] = s
        res['f'] = np.sqrt(f)  # for plotting data
        res['p_norm'] = p_norm  # for plotting data
        res['stepsize'] = stepsize  # for plotting data
        res['theta'] = theta  # for plotting data
        return res


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    # plot Fig. 2
    ndim_problem = 200
    problem = {'fitness_function': cigar,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'fitness_threshold': 1e-8,
               'seed_rng': 2022,  # not given in the original paper
               'sigma': 20 / 3}
    r1es = Fig2(problem, options)
    results = r1es.optimize()
    x = np.arange(len(results['f'])) + 1
    plt.figure()
    plt.yscale('log')
    plt.plot(x, results['f'], 'b-', label='sqrt(f)', fillstyle='none')
    plt.plot(x, results['p_norm'], 'g-', label='||p||', fillstyle='none')
    plt.plot(x, results['stepsize'], 'k-', label='sigma', fillstyle='none')
    plt.plot(x, results['theta'], 'r-', label='theta', fillstyle='none')
    plt.xticks(ticks=[1, 1000, 2000, 3000])
    plt.xlim(1, 3000)
    plt.yticks(ticks=[1e-9, 1e-6, 1e-3, 1e0, 1e3, 1e6])
    plt.xlabel('Generations')
    plt.ylabel('Values')
    plt.legend()
    plt.show()
    # plot Fig. 3
    ndim_problem = 1000
    for func in [ellipsoid, discus, rosenbrock]:
        problem = {'fitness_function': func,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -10 * np.ones((ndim_problem,)),
                   'upper_boundary': 10 * np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-8,
                   'max_function_evaluations': 1e8,
                   'seed_rng': 2022,  # not given in the original paper
                   'sigma': 20 / 3,
                   'saving_fitness': 1,
                   'is_restart': False}
        r1es = Fig2(problem, options)
        results = r1es.optimize()
        pickle.dump(results, open(func.__name__ + '_' + str(ndim_problem) + '.pickle', 'wb'))
        plot(func.__name__, ndim_problem)



from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.res import RES
from pypop7.optimizers.es.ssaes import SSAES
from pypop7.optimizers.es.dsaes import DSAES
from pypop7.optimizers.es.csaes import CSAES
from pypop7.optimizers.es.saes import SAES
from pypop7.optimizers.es.opoc2006 import OPOC2006
from pypop7.optimizers.es.sepcmaes import SEPCMAES
from pypop7.optimizers.es.opoc2009 import OPOC2009
from pypop7.optimizers.es.ccmaes2009 import CCMAES2009
from pypop7.optimizers.es.opoa2010 import OPOA2010
from pypop7.optimizers.es.opoa2015 import OPOA2015
from pypop7.optimizers.es.ccmaes2016 import CCMAES2016
from pypop7.optimizers.es.maes import MAES
from pypop7.optimizers.es.fmaes import FMAES
from pypop7.optimizers.es.r1es import R1ES
from pypop7.optimizers.es.rmes import RMES
from pypop7.optimizers.es.lmcmaes import LMCMAES
from pypop7.optimizers.es.lmcma import LMCMA
from pypop7.optimizers.es.lmmaes import LMMAES
from pypop7.optimizers.es.mmes import MMES


__all__ = [ES,  # base (abstract) class
           RES, SSAES, DSAES, CSAES, SAES,  # representative ES versions during development
           OPOC2006, OPOC2009, CCMAES2009, OPOA2010, OPOA2015, CCMAES2016, MAES, FMAES,  # modern ES versions
           SEPCMAES, R1ES, RMES, LMCMAES, LMCMA, LMMAES, MMES]  # especially for large-scale black-box optimization



"""Repeat Fig. 4 Fast CMA line from the following paper:
    Z. Li, Q. Zhang, X. Lin, H. zhen
    Fast Covariance Matrix Adaptation for Large-Scale Black-Box Optimization
    IEEE Transaction on Cybernetics vol.50 No.5 May 2020
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=8533604
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle

from pypop7.optimizers.es.fcmaes import FCMAES as Solver
from benchmarks.base_functions import cigar, rosenbrock, discus, ellipsoid, different_powers, _squeeze_and_check

sns.set_theme(style='darkgrid')


def ctb(x):
    x = np.power(_squeeze_and_check(x, True), 2)
    y = x[0] + 1e6 * x[-1]
    for i in range(1, len(x)-1):
        y += 1e4 * x[i]
    return y


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function):
    plt.figure()
    result1 = read_pickle(function, 512)
    result2 = read_pickle(function, 1024)
    plt.yscale('log')
    plt.ticklabel_format(axis="x", style="sci", scilimits=(0, 0))
    if function == 'cigar' or function == 'ctb':
        plt.ylim([1e-10, 1e10])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5, 1e10])
        plt.xlim([0, 5e5])
        plt.xticks([0, 1e5, 2e5, 3e5, 4e5, 5e5])
    elif function == 'ellipsoid':
        plt.ylim([1e-10, 1e10])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5, 1e10])
        plt.xlim([0, 2.5e7])
        plt.xticks([0, 0.5e7, 1e7, 1.5e7, 2e7, 2.5e7])
    elif function == 'rosenbrock':
        plt.ylim([1e-10, 1e5])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 5e7])
        plt.xticks([0, 1e7, 2e7, 3e7, 4e7, 5e7])
    elif function == 'discus':
        plt.ylim([1e-10, 1e5])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 1e7])
        plt.xticks([0, 2e6, 4e6, 6e6, 8e6, 1e7])
    elif function == 'different_powers':
        plt.ylim([1e-10, 1e5])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 6e6])
        plt.xticks([0, 1e6, 2e6, 3e6, 4e6, 5e6])
    mark_on1 = [int(len(result1['fitness']) / 3), int(len(result1['fitness']) / 3 * 2), len(result1['fitness'])-1]
    mark_on2 = [int(len(result2['fitness']) / 3), int(len(result2['fitness']) / 3 * 2), len(result2['fitness'])-1]
    plt.plot(result1['fitness'][:, 0], result1['fitness'][:, 1], color='r', marker='s',
             markevery=mark_on1, markerfacecolor='none', label='n=512')
    plt.plot(result2['fitness'][:, 0], result2['fitness'][:, 1], color='r', marker='o',
             markevery=mark_on2, markerfacecolor='none', label='n=1024')
    plt.xlabel("evaluations")
    plt.ylabel("objective values")
    plt.title(function.capitalize())
    plt.legend(labels=['n=512', 'n=1024'], loc='best')
    plt.savefig('result_' + function + '.png')
    plt.show()


if __name__ == '__main__':
    # plot Fig. 4
    for f in [cigar, ellipsoid, rosenbrock, discus, ctb, different_powers]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        for d in [512, 1024]:
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -10 * np.ones((d,)),
                       'upper_boundary': 10 * np.ones((d,))}
            options = {'fitness_threshold': 1e-8,
                       'max_function_evaluations': 3e7,
                       'seed_rng': 0,
                       'sigma': 0.1,
                       'verbose_frequency': 2e3,
                       'record_fitness': True,
                       'record_fitness_frequency': 1}
            if f.__name__ == 'cigar':
                options['max_function_evaluations'] = 2e5
            elif f.__name__ == 'ellipsoid':
                options['max_function_evaluations'] = 2e7
            elif f.__name__ == 'rosenbrock':
                options['max_function_evaluations'] = 1.5e7
            elif f.__name__ == 'discus':
                options['max_function_evaluations'] = 4e6
            elif f.__name__ == 'ctb':
                options['max_function_evaluations'] = 2e5
            elif f.__name__ == 'different_powers':
                options['max_function_evaluations'] = 1e6
            solver = Solver(problem, options)
            results = solver.optimize()
            write_pickle(f.__name__, d, results)
        plot(f.__name__)



from pypop7.optimizers.cem.cem import CEM
from pypop7.optimizers.cem.scem import SCEM
from pypop7.optimizers.cem.dscem import DSCEM
from pypop7.optimizers.cem.dcem import DCEM


__all__ = [CEM,  # base (abstract) class
           SCEM, DSCEM, DCEM]



import numpy as np
import matplotlib.pyplot as plt


# helper function
def generate_xyz(func, x, y, num=200):
    x, y = np.array(x), np.array(y)
    if x.size == 2:
        x = np.linspace(x[0], x[1], num)
    if y.size == 2:
        y = np.linspace(y[0], y[1], num)
    x, y = np.meshgrid(x, y)
    z = np.empty(x.shape)
    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            z[i, j] = func([x[i, j], y[i, j]])
    return x, y, z


def plot_contour(func, x, y, levels=None, num=200, is_save=False):
    x, y, z = generate_xyz(func, x, y, num)
    if levels is None:
        plt.contourf(x, y, z, cmap='cool')
        plt.contour(x, y, z, colors='white')
    else:
        plt.contourf(x, y, z, levels, cmap='cool')
        c = plt.contour(x, y, z, levels, colors='white')
        plt.clabel(c, inline=True, fontsize=12, colors='white')
    plt.title(func.__name__)
    plt.xlabel('x')
    plt.ylabel('y')
    if is_save:
        plt.savefig(func.__name__ + '_contour.png')
    plt.show()
